#include <cstdio>

#include <cstdlib>

#include <vector>
#include <cmath>

#include "rlsem.hpp"

#include "em.hpp"

#include "kmeans.hpp"



static bool SwapRandom(MixtureModel& MM, const Data& D) {


	double MMCount = MM.Count();
	double DCount = D.Count();
	int ComponentIndex = int( (MMCount * rand()) / RAND_MAX );

	if (ComponentIndex == MMCount) ComponentIndex = MMCount - 1;

    int PointIndex = int( (DCount * rand()) / RAND_MAX );

    if (PointIndex == DCount ) PointIndex = DCount - 1;

    for (int d = 0; d < MM.Dimensionality(); ++d)

        MM[ComponentIndex].Mean()[d] = D(PointIndex, d);

//	FILE* f = fopen("swap2.txt", "a+");
//	MM.Write(f, MM_FF_PLAIN);
//	fprintf(f, "models after swap\n\n");
	return true; 

}





static int SwapRandom2(MixtureModel& MM, const Data& D) {

	double MMCount = MM.Count();
	double DCount = D.Count();

	int ComponentIndex = int( (MMCount * rand()) / RAND_MAX );

	if (ComponentIndex == MMCount) ComponentIndex = MMCount - 1;

    int PointIndex = int( (DCount * rand()) / RAND_MAX );

    if (PointIndex == DCount ) PointIndex = DCount - 1;

    for (int d = 0; d < MM.Dimensionality(); ++d)

        MM[ComponentIndex].Mean()[d] = D(PointIndex, d);

    return ComponentIndex;

}





static bool SwapPostPartitioning(MixtureModel& MM, const Data& D) {

    if (!SwapRandom(MM, D)) return false;

    MixtureModel copy(MM);

    KMeans(D, copy, 1);

    for (int k = 0; k < MM.Count(); ++k) MM[k].Weight(copy[k].Weight());

/*
	FILE* f = fopen("swap1.txt", "a+");
	MM.Write(f, MM_FF_PLAIN);
	fprintf(f, "models after swap\n\n");*/
    return true;

}



static bool SwapPartitionSwapped(MixtureModel& MM, const Data& D) {

  int ComponentIndex = SwapRandom2(MM, D);

  MixtureModel copy(MM);

  KMeans(D, copy, 1);

  MM[ComponentIndex].Weight(copy[ComponentIndex].Weight());
/*
	FILE* f = fopen("swap4.txt", "a+");
	MM.Write(f, MM_FF_PLAIN);
	fprintf(f, "models after swap\n\n");*/

  return true;

}



static bool SwapEraseCreate(MixtureModel& MM, const Data& D) {
	
	double MMCount = MM.Count();
	double DCount = D.Count();
/*
	FILE* f = fopen("swap.txt", "a+");
	fprintf(f, "\n models before RS\n");
	MM.Write(f, MM_FF_PLAIN);*/
	

    // Erase a component by setting the weight to zero.
	int ComponentIndex = int( (MMCount * rand()) / RAND_MAX );

	if (ComponentIndex == MMCount) ComponentIndex = MMCount - 1;

    for (int d = 0; d < MM.Dimensionality(); ++d)

        MM[ComponentIndex].Mean()[d] = 0;

    MM[ComponentIndex].Weight(0.0);
	
//	fprintf(f, "\n Remove component %d\n", ComponentIndex);
//	MM.Write(f, MM_FF_PLAIN);

    // Change the mean vector of erased component to random position.
    int PointIndex = int( (DCount * rand()) / RAND_MAX );

    if (PointIndex == DCount ) PointIndex = DCount - 1;

    for (int d = 0; d < MM.Dimensionality(); ++d)

        MM[ComponentIndex].Mean()[d] = D(PointIndex, d);

    // Make a weighted sum of covariance matrices of other components by using

    // the probabilities of the new center point as weights.
//	fprintf(f, "\n add component\n");
//	MM.Write(f, MM_FF_PLAIN);

    MM.NormalizeWeights();


    // Compute probability of new center point wrt. other components.

    std::vector<std::pair<double,Matrix> > DetInvs;

    SetupDeterminantAndInverseMatrices(DetInvs, MM);

    Matrix PosteriorRow(1, MM.Count(), 0.0);

    ConstRowIterator VectorBegin = MM[ComponentIndex].Mean().ConstRowBegin(0);

    //    ConstRowIterator VectorEnd = MM[ComponentIndex].Mean().ConstRowEnd(0);

    double Denom = 0;

    for (int k = 0; k < MM.Count(); ++k) if (k != ComponentIndex) {

        double Mah2;

        if (MM[k].MatrixType() == MM_COV_TYPE_DIAG)

            Mah2 = MahalanobisDiag2(VectorBegin,

                MM[k].Mean().ConstRowBegin(0), DetInvs[k].second);

        else Mah2 = MahalanobisChol2(VectorBegin,

            MM[k].Mean().ConstRowBegin(0), DetInvs[k].second);

        PosteriorRow[k] =

            MM[k].Weight() * ProbabilityDensity(Mah2, DetInvs[k].first);

        Denom += PosteriorRow[k];

    }

    if (Denom > DBL_MIN) PosteriorRow *= 1.0 / Denom;

    else {

        PosteriorRow.Fill(1.0 / (MM.Count() - 1));

        PosteriorRow[ComponentIndex] = 0.0;

    }

    // Weighted sum.

    MM[ComponentIndex].Covariance().Fill(0.0);

    double W = 0.0;

    for (int k = 0; k < MM.Count(); ++k) if (k != ComponentIndex) {

        MM[ComponentIndex].Covariance() += PosteriorRow[k] * MM[k].Covariance();

        W += PosteriorRow[k] * MM[k].Weight();

    }

    MM[ComponentIndex].Weight(W);

    MM.NormalizeWeights();

    // Perhaps this would be a more sensible weight update.

    /*

    MixtureModel copy(MM);

    KMeans(D, copy, 1);

    for (int k = 0; k < MM.Count(); ++k) MM[k].Weight(copy[k].Weight());

    */
//	fprintf(f, "\n update parameters (after RS)\n");
//	MM.Write(f, MM_FF_PLAIN);


    return true;

}





static bool Swap(MixtureModel& MM, const Data& D, RLSEMSwapType SwapType) {

    switch(SwapType) {

    case RLSEMSwap_Random: return SwapRandom(MM, D);

    case RLSEMSwap_PostPartitioning: return SwapPostPartitioning(MM, D);

    case RLSEMSwap_EraseCreate: return SwapEraseCreate(MM, D);

    case RLSEMSwap_PartitionSwapped: return SwapPartitionSwapped(MM, D); 

    }

    fprintf(stderr, "RLSEM: Unknown swap type: %i.\n", int(SwapType));

    return false;

}





double RLSExpectationMaximization(MixtureModel& MM, const Data& D,

    const Matrix& MinimumVariances, int RLSIterations, int EMIterations,

    RLSEMSwapType SwapType, bool Verbose)

{
//	FILE* f = fopen("RSEM.txt", "a+");
    MixtureModel& BestModel = MM;

    double BestLL = emLogLikelihood(D, MM);

    for (int k = 0; k < RLSIterations; ++k) {

        MixtureModel Model = BestModel;

        if (!Swap(Model, D, SwapType)) {

            fprintf(stderr,

                "RLSExpectationMaximization: problem during swap.\n");

            return 0.0;

        }
			
//		fprintf(stdout, "RS iteration = %d\n", k+1);

        double LL =

            ExpectationMaximization(Model, D, MinimumVariances, EMIterations, Verbose);
//		fprintf(stdout, "%f\n", LL);

        if (LL > BestLL && !emNumericalProblems()) {
//			fprintf(stdout, "0\n");
            BestLL = LL;

            BestModel = Model;

//			fprintf(f, "\n----models after %d success RS iteration\n\n", k+1);
//			BestModel.Write(f, MM_FF_PLAIN);
			
			
        }
	
	if (Verbose) printf("%3i : %g\n", k + 1, BestLL);

    }
//	printf("%g\n",  BestLL);
    return BestLL;

}


double RLSExpectationMaximization(MixtureModel& MM, const Data& D,

    const Matrix& MinimumVariances, int RLSIterations, double Threshold,

    RLSEMSwapType SwapType, bool Verbose)

{
//	FILE* f = fopen("RSEM.txt", "a+");
    MixtureModel& BestModel = MM;
   
	double BestLL = emLogLikelihood(D, MM);
    
	for (int k = 0; k < RLSIterations; ++k) {
//		fprintf(f, "\n iter# %d\n\n", k);

        MixtureModel Model = BestModel;
//		fprintf(f, "before swap\n\n");
//		Model.Write(f, MM_FF_PLAIN);

        if (!Swap(Model, D, SwapType)) {

            fprintf(stderr,

                "RLSExpectationMaximization: problem during swap.\n");

            return 0.0;

        }
		int temp_iter = 0;
        double LL =

            ExpectationMaximization(Model, D, MinimumVariances, Threshold, Verbose, temp_iter);
//			fprintf(f, "\n after EM\n");
//			Model.Write(f, MM_FF_PLAIN);

        if (LL > BestLL && !emNumericalProblems()) {
//			fprintf(stdout, "0\n");
            BestLL = LL;

            BestModel = Model;

//			fprintf(f, "\n wa, it's improved!\n");
//			fprintf(stdout, "it's improved on iter %d!\n", k);
//			fprintf(f, "\n----models after %d success RS iteration\n\n", k+1);
//			BestModel.Write(f, MM_FF_PLAIN);
			
			
        }
	
	if (Verbose) printf("%3i : %g\n", k + 1, BestLL);

    }
    return BestLL;

}





